w
import os, sys, pickle, bisect, random, math, time, collections

DEBUG = False
DISSING = True
CHECK_ASSUMPTIONS_CAREFULLY = False
DUMP_BAD_SOLUTIONS = False
BIG_NUMBER = 1 << 64
GREYS = ["white", "gray55", "gray70", "gray90"]
rt1 = []
rt2 = []
rt3 = []
rt4 = []
srec = []
log = None
mumu = 0
klf = None
bloc = "."
COLORS = [
    "bisque3",
    "brown",
    "chartreuse",
    "coral",
    "cyan3",
    "darkorchid1",
    "gold",
    "burlywood1",
    "navajowhite1",
    "oldlace",
    "hotpink",
    ]

class UnsupportedMode(Exception):
    pass

class Tree_Maker:
    def __init__(self):
        record = pickle.load(file("rep_list.dat", "rb"))
        self.method_sizes = record.get("method_sizes")
        self.edge_weights = record.get("edge_weights")
        self.out_degrees = record.get("out_degrees")

        
    def generate(self, spm_size, tree_size, back_edge_probability, r):
        upper_limit = bisect.bisect_right(self.method_sizes, spm_size)
        assert upper_limit > 0
       
        node_size = [r.choice(self.method_sizes[:upper_limit])]
        num_children = [r.choice(self.out_degrees)]
        edges = [[]]

        while len(node_size) <= tree_size:
            # Find node without enough children
            parents = []
            for parent in xrange(len(node_size)):
                #assert len(edges[parent]) <= num_children[parent]
                if len(edges[parent]) < num_children[parent]:
                    parents.append(parent)

            if len(parents) == 0:
                # Consider any parent
                parents = range(len(node_size))
           
            # Get parent number
            parent = r.choice(parents)
            child = len(node_size)

            # add new child node
            node_size.append(r.choice(self.method_sizes[:upper_limit]))
            num_children.append(r.choice(self.out_degrees))
            edges.append([])
            edges[parent].append(child)

        # Rebuild edge list as dictionary; add weights
        edge_dict = dict()
        for parent in xrange(len(node_size)):
            for child in edges[parent]:
                assert child > parent
                edge_dict[(parent, child)] = r.choice(self.edge_weights)


        if back_edge_probability <= 0:
            return Tree(node_size, edge_dict, [])

        # Add back edges
        depth = [None for i in xrange(len(node_size))]
        parent = [None for i in xrange(len(node_size))]
        depth[0] = 0
        change = True
        while change:
            change = False
            for x in xrange(len(node_size)):
                for y in edges[x]:
                    if depth[y] == None:
                        depth[y] = depth[x] + 1
                        parent[y] = x
                        change = True

        back_edges = []
        for footer in xrange(1, len(node_size)):
            fdepth = depth[footer] 
            assert fdepth > 0
            if len(edges[footer]) != 0:
                continue
            if r.random() >= back_edge_probability:
                continue

            hdepth = r.randrange(0, fdepth)
            assert 0 <= hdepth < fdepth
            header = footer
            for i in xrange(fdepth - hdepth):
                header = parent[header]
                assert header != None

            weight = edge_dict[(parent[footer], footer)]
            back_edges.append((header, footer, weight))

        return Tree(node_size, edge_dict, back_edges)

class Tree:
    def __init__(self, nodes, edges, back_edges):
        self.nodes = nodes
        self.edges = edges
        self.back_edges = back_edges
        self.recompute_derived_data()
        self.quantum = 4
        self.fixed_cost = 0

    def write_csearch(self, fd, spm_size, mode):
        fd.write("%u %u %s %u %u\n" % (
                spm_size, len(self.nodes), mode, 
                self.quantum, self.fixed_cost))

        for i in xrange(len(self.nodes)):
            p = self.parent[i]
            if p == None:
                assert i == 0
                p = 0
            else:
                assert p < i

            fd.write("%-8u %-8u %-8u 0\n" % (
                self.nodes[i],
                p, 
                self.edges.get((p, i), 1)))

        if mode in "wW":
            fd.write("%u\n" % len(self.back_edges))
            for i in xrange(len(self.back_edges)):
                (header, footer, weight) = self.back_edges[i]
                fd.write("%-8u %-8u %-8u\n" % (header, footer, weight))
           
        fd.write("0\n")

    def read_csearch(self, fd):
        region_contents = dict()
        mode = None
        spm_size = None
        max_region_number = 0

        # header
        fields = fd.readline().split()
        assert len(fields) == 5
        spm_size = int(fields[0])
        assert len(self.nodes) == int(fields[1])
        mode = fields[2]
        assert self.quantum == int(fields[3])
        assert self.fixed_cost == int(fields[4])

        # nodes
        for node_number in xrange(len(self.nodes)):
            fields = fd.readline().split()
            assert len(fields) == 4
            assert self.nodes[node_number] == int(fields[0]) 
            p = int(fields[1]) 
            if node_number == 0:
                assert p == 0
                p = None
            else:
                assert self.parent[node_number] == p

            assert self.edges.get((p, node_number), 1) == int(fields[2]), line
            region_number = int(fields[3])

            if region_contents.has_key(region_number):
                region_contents[region_number].append(node_number)
            else:
                region_contents[region_number] = [node_number]

            max_region_number = max(region_number, max_region_number)

        # back edges
        if mode in "wW":
            fields = fd.readline().split()
            assert len(fields) == 1
            p = int(fields[0])
            assert p == len(self.back_edges)

            for i in xrange(len(self.back_edges)):
                fields = fd.readline().split()
                assert len(fields) == 3
                header = int(fields[0])
                footer = int(fields[1])
                weight = int(fields[2])
                assert (header, footer, weight) == self.back_edges[i]

        partition = []
        assert mode != None
        assert spm_size != None
        for region_number in xrange(0, max_region_number + 1):
            assert region_contents.has_key(region_number)
            partition.append(region_contents[region_number])

        self.fixup(partition)
        return (partition, spm_size, mode)

    def recompute_derived_data(self):
        self.edges2 = self.edges.items()
        self.edges2.sort()

        self.parent = [ 0 for i in self.nodes ]
        for ((x, y), cost) in self.edges2:
            assert self.parent[y] == 0
            assert 0 <= x < y
            self.parent[y] = x

        self.children = [[] for i in xrange(len(self.nodes))]
        for ((x, y), cost) in self.edges2:
            self.children[x].append(y)

        assert self.parent[0] == 0
        self.parent[0] = None

        self.depth = [None for i in xrange(len(self.nodes))]
        self.depth[0] = 0
        change = True
        while change:
            change = False
            for ((x, y), cost) in self.edges2:
                if ((self.depth[x] != None)
                and (self.depth[y] == None)):
                    self.depth[y] = self.depth[x] + 1
                    change = True

    def remove_subtree(self, subtree_root):
        nodes_to_remove = []
        def whole_subtree(x0):
            for ((x, y), cost) in self.edges2:
                if x == x0:
                    whole_subtree(y)

            nodes_to_remove.append(x0)

        whole_subtree(subtree_root)
        nodes_to_remove.sort()
        self.renumber(nodes_to_remove)

    def renumber(self, nodes_to_remove, breadth_first = False):
        renumber = range(len(self.nodes))
        for i in reversed(nodes_to_remove):
            self.nodes.pop(i)
            renumber[i] = None

        j = 0
        if not breadth_first:
            for i in xrange(len(renumber)):
                if renumber[i] != None:
                    renumber[i] = j
                    j += 1
        else:
            for i in xrange(len(renumber)):
                if renumber[i] != None:
                    renumber[i] = -1

            for d in xrange(0, 1 + max(self.depth)):
                for i in xrange(len(renumber)):
                    if (renumber[i] == -1) and (self.depth[i] == d):
                        renumber[i] = j
                        j += 1

        assert j == len(self.nodes), (j, renumber)

        self.edges = dict()
        for ((x, y), cost) in self.edges2:
            x = renumber[x]
            y = renumber[y]
            if (x != None) and (y != None) and (cost > 0):
                self.edges[(x, y)] = cost

        nbe = []
        for (x, y, cost) in self.back_edges:
            x = renumber[x]
            y = renumber[y]
            if (x != None) and (y != None) and (cost > 0):
                nbe.append((x, y, cost))

        self.back_edges = nbe
        self.recompute_derived_data()

    def swap_children(self, m1, m2):
        assert m1 != m2
        assert self.parent[m1] == self.parent[m2]
        assert self.parent[m1] != None

        # swap weights
        (self.nodes[m1], self.nodes[m2]) = (self.nodes[m2], self.nodes[m1])

        # swap edges
        self.edges = dict()
        for ((x, y), cost) in self.edges2:
            if x == m1:
                x = m2
            elif x == m2:
                x = m1

            if y == m1:
                y = m2
            elif y == m2:
                y = m1

            self.edges[(x, y)] = cost

        self.recompute_derived_data()

    def dump(self):
        return "Tree(%s, %s, %s)\n" % (
                repr(self.nodes), repr(self.edges),
                repr(self.back_edges))

    def check_partition(self, partition):
        N = len(self.nodes)
        assert type(partition) == list
        assert len(partition) != 0
        for region in partition:
            assert type(region) == list
            assert len(region) != 0
            for member in region:
                assert type(member) == int
                assert 0 <= member < N

    def graphviz(self, partition = None):
        c = dict()
        if partition != None:
            #for region in partition:
            #    i = min(region)
            #    for member in region:
            #        c[member] = COLORS[i % len(COLORS)]
            d = self.partition_to_dict(partition)
            c[0] = 0
            for y in xrange(1, len(self.nodes)):
                x = self.parent[y]
                if d.get(x, -1) == d.get(y, -1):
                    c[y] = c[x]
                else:
                    c[y] = (c[x] + 1) % len(GREYS)


        out = []
        out.append("digraph G {")
        shape = "ellipse"
        for (i, size) in enumerate(self.nodes):
            colour = GREYS[c.get(i, 0)]
            
            out.append("""
node [shape=%(shape)s, style=filled, fillcolor=%(colour)s, 
    label="m%(nodenumber)u, size %(size)u"] N%(nodenumber)u;
""" % { "nodenumber" : i,
        "colour" : colour,
        "size" : size,
        "shape" : shape,
        })
            shape = "box"

        for ((x, y), weight) in self.edges2:
            out.append('N%u -> N%u [label="%u"];' % (x, y, weight))

        for (x, y, weight) in self.back_edges:
            out.append('N%u -> N%u [label="%u !"];' % (y, x, weight))

        out.append("}")
        out.append("")
        return '\n'.join(out)

    def quantise(self, q):
        self.nodes = [ ((n + q - 1) / q) * q for n in self.nodes ]
        self.quantum = q

    def get_value_w(self, partition):
        # edge weights * region sizes only
        d = self.partition_to_dict(partition)
        region_weight = dict()

        for region in partition:
            Rw = self.get_region_weight(region)
            for member in region:
                region_weight[member] = Rw

        total = 0
        for ((x, y), weight) in self.edges2 + [((-1, 0), 1)]:
            if ((d.get(x, -1) != d.get(y, -1))
            and region_weight.has_key(y)):      
                total += (region_weight[y] + self.fixed_cost) * weight

        for (x, y, weight) in self.back_edges:
            if ((d.get(x, -1) != d.get(y, -1))
            and region_weight.has_key(x)):      
                total += (region_weight[x] + self.fixed_cost) * weight

        return total

    def get_value_prime(self, partition):
        # edge weights only
        d = self.partition_to_dict(partition)
        total = 0
        for ((x, y), weight) in self.edges2:
            dx = d.get(x, -1) 
            if dx != d.get(y, dx):
                total += weight

        return total

    def get_value_gamma(self, partition):
        # edge weights * region sizes only
        d = self.partition_to_dict(partition)
        region_weight = dict()

        for region in partition:
            Rw = self.get_region_weight(region)
            for member in region:
                region_weight[member] = Rw

        total = 0
        for ((x, y), weight) in self.edges2:
            if ((d.get(x, -1) != d.get(y, -1))
            and region_weight.has_key(y)):      

                total += (region_weight.get(x, 0) + self.fixed_cost) * weight
                total += (region_weight[y] + self.fixed_cost) * weight

        total += (region_weight.get(0, 0) + self.fixed_cost)

        return total

    def partition_to_dict(self, partition):
        assert type(partition) == list
        location = dict()
        for region in partition:
            assert type(region) == list
            assert len(region) > 0
            for member in region:
                location[member] = region

        return location

    def get_region_weight(self, region):
        region_weight = 0
        for member in region:
            region_weight += self.nodes[member]
        return region_weight

    def fixup(self, partition):
        for r in partition:
            r.sort()
        partition.sort()

    def get_weight(self, partition):
        # Get the weight of this partition
        assert type(partition) == list
        partition_weight = 0
        for region in partition:
            assert type(region) == list
            assert len(region) > 0
            region_weight = self.get_region_weight(region)
            partition_weight = max(partition_weight, region_weight)

        return partition_weight

    def get_value(self, partition, mode):
        return {
                "C" : self.get_value_gamma,
                "W" : self.get_value_w,
                "w" : self.get_value_w,
                "P" : self.get_value_prime,
                }[mode](partition)

        return partition

class Tree_Maker_2(Tree_Maker):
    def generate(self, spm_size, tree_size, back_edge_probability, r):
        t1 = Tree_Maker.generate(self, spm_size, tree_size,
                back_edge_probability, r)

        upper_limit = bisect.bisect_right(self.method_sizes, spm_size)
        assert upper_limit > 0

        t1.edges = dict()
        for ((x, y), _) in t1.edges2:
            t1.edges[(x, y)] = 0

        is_ok = set()
        be_count = dict()
        be_target = dict()
        for (x, y, _) in t1.back_edges:
            assert not be_target.has_key(y)
            be_target[y] = x
            be_count[(x, y)] = 0

        for i in xrange(4):
            x = 0
            is_ok.add(x)
            for j in xrange(1000):
                is_ok.add(x)
                if be_target.has_key(x) and r.randint(1, 3) == 1:
                    y = x
                    x = be_target[y]
                    be_count[(x, y)] += 1

                elif len(t1.children[x]) == 0:
                    break

                else:
                    y = r.choice(t1.children[x])
                    t1.edges[(x, y)] += 1
                    x = y

            while be_target.has_key(x):
                is_ok.add(x)
                y = x
                x = be_target[y]
                be_count[(x, y)] += 1

            is_ok.add(x)
            if len(t1.children[x]) == 0:
                # didn't reach the end? create new node
                y = len(t1.nodes)
                is_ok.add(y)
                t1.nodes.append(r.choice(self.method_sizes[:upper_limit]))
                t1.edges[(x, y)] = 1
                t1.recompute_derived_data()

        t1.back_edges = []
        for (y, x) in be_target.iteritems():
            v = be_count[(x, y)]
            if v != 0:
                t1.back_edges.append((x, y, v))

        t1.back_edges.sort()
        nodes_to_remove = []

        #for (x, y) in t1.edges.keys():
        #    if t1.edges[(x, y)] == 0:
        #        nodes_to_remove.append(y)
        #        del t1.edges[(x, y)]

        nodes_to_remove = list(set(range(len(t1.nodes))) - is_ok)
        nodes_to_remove.sort()
        print 'remove', nodes_to_remove
        t1.renumber(nodes_to_remove)

        for (x, y) in t1.edges.keys():
            assert t1.edges[(x, y)] != 0

        return Tree(t1.nodes, t1.edges, t1.back_edges)

